#include "Module_builtins.h"
#include "Object_Custom.h"
#include "Object_List.h"
#include "Object_Struct.h"
#include "Object_Iterator.h"
#include "Object_Function.h"

AScript_BeginModule(__builtins__)

// declaration of symbols privately used in this module
AScript_DeclarePrivSymbol(open);
AScript_DeclarePrivSymbol(open_l);
AScript_DeclarePrivSymbol(open_r);

//-----------------------------------------------------------------------------
// AScript module functions: __builtins__
//-----------------------------------------------------------------------------
// num = int(num):map
AScript_DeclareFunctionEx(int_, "int")
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType);
}

AScript_ImplementFunction(int_)
{
	const Value &value = context.GetValue(0);
	Value result;
	if (value.IsNumber()) {
		result.SetNumber(static_cast<long>(value.GetNumber()));
	} else if (value.IsComplex()) {
		result.SetNumber(static_cast<long>(std::abs(value.GetComplex())));
	} else if (value.IsExprOrSymbol()) {
		result.InitAsExpr(env, new Expr_Caller(this, new Expr_Value(value)));
	} else if (value.IsValid()) {
		SetError_InvalidValType(sig, value);
	}
	return result;
}

// func = class(super_class?) {block?}
AScript_DeclareFunctionEx(class_, "class")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "super_class", VTYPE_Function, false, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(static_));
}

AScript_ImplementFunction(class_)
{
	const ValueList &valListArg = context.GetArgs();
	const Expr_Block *pExprBlock = context.GetBlock();
	Class *pClassSuper = env.GetClass_Object();
	const Value &valueSuper = context.GetValue(0);
	if (valueSuper.IsFunction()) {
		pClassSuper = valueSuper.GetFunction()->GetClassToConstruct();
		if (pClassSuper == NULL) {
			valueSuper.GetFunction()->SetError_NotConstructor(sig);
			return Value::Null;
		}
	}
	Class_Custom *pClassCustom = new Class_Custom(pClassSuper, AScript_Symbol(_anonymous_));
	//pClassCustom->AssignValue(AScript_Symbol(super), valueSuper, false); // *****
	if (pExprBlock != NULL && !pClassCustom->BuildContent(env, sig, pExprBlock)) {
		Class::Delete(pClassCustom);
		return Value::Null;
	}
	FunctionCustom *pFuncInit = dynamic_cast<FunctionCustom *>(
						pClassCustom->LookupFunction(AScript_Symbol(__init__), false));
	Value result;
	FunctionType funcType = context.IsSet(AScript_Symbol(static_))?
											FUNCTYPE_Class : FUNCTYPE_Function;
	if (pFuncInit != NULL) {
		FunctionCustom *pFunc = new FunctionCustom(*pFuncInit, env);
		pFunc->SetSymbol(AScript_Symbol(_anonymous_));
		pFunc->SetClassToConstruct(pClassCustom);
		pClassCustom->SetConstructor(pFunc);
		result.InitAsFunction(env, pFunc);
	} else {
		FunctionCustom *pFunc = new FunctionCustom(env,
			AScript_Symbol(_anonymous_), NULL, pClassCustom, funcType);
		ContextExpr contextExpr(ExprList::Null);
		if (!pFunc->CustomDeclare(env, sig, SymbolSet::Null, contextExpr)) {
			Function::Delete(pFunc);
			return Value::Null;
		}
		result.InitAsFunction(env, pFunc);
	}
	return result;
}

// func = struct(`args*) {block?}
// if :loose attribute is specified, arguments in the generated function
// will get the following modification.
// - Once attribute will be modified to ZeroOrOnce.
// - OnceOrMore attribute will be modified to ZeroOrMore
class Func_struct : public Function {
public:
	class Constructor : public ConstructorBase {
	public:
		inline Constructor(Class *pClass);
		virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	};
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_struct(Environment &env, const char *name = "struct");
};
Func_struct::Func_struct(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "args", VTYPE_Quote, false, OCCUR_OnceOrMore);
	DeclareBlock(OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(loose));
}

Value Func_struct::DoEval(Environment &env, Signal sig, Context &context) const
{
	Class *pClassSuper = env.GetClass_Struct();
	ExprList exprArgs;
	foreach_const (ValueList, pValue, context.GetList(0)) {
		exprArgs.push_back(const_cast<Expr *>(pValue->GetExpr()));
	}
	Value result;
	Class_Custom *pClassCustom =
				new Class_Custom(pClassSuper, AScript_Symbol(_anonymous_));
	const Expr_Block *pExprBlock = context.GetBlock();
	if (pExprBlock != NULL && !pClassCustom->BuildContent(env, sig, pExprBlock)) {
		return Value::Null;
	}
	Constructor *pFunc = new Constructor(pClassCustom);
	pClassCustom->SetConstructor(pFunc->IncRef());
	ContextExpr contextExpr(context.GetAttrs(), SymbolSet::Null, NULL, NULL,
														Value::Null, exprArgs);
	if (!pFunc->CustomDeclare(env, sig, _attrsOpt, contextExpr)) {
		Function::Delete(pFunc);
		return false;
	}
	if (context.IsSet(AScript_Symbol(loose))) {
		pFunc->GetDeclList().SetAsLoose();
	}
	result.InitAsFunction(env, pFunc);
	return result;
}

Func_struct::Constructor::Constructor(Class *pClass) :
							ConstructorBase(AScript_Symbol(_anonymous_), pClass)
{
}

Value Func_struct::Constructor::DoEval(Environment &env, Signal sig, Context &context) const
{
	Object *pObj = GetClassToConstruct()->
						CreateDescendant(env, sig, GetClassToConstruct());
	ValueList::const_iterator pValue = context.GetArgs().begin();
	DeclarationList::const_iterator ppDecl = GetDeclList().begin();
	for ( ; pValue != context.GetArgs().end() && ppDecl != GetDeclList().end();
														pValue++, ppDecl++) {
		pObj->AssignValue((*ppDecl)->GetSymbol(), *pValue, false);
	}
	return Value(pObj, VTYPE_Object);
}

// val = module {block}
AScript_DeclareFunction(module)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_Once);
}

AScript_ImplementFunction(module)
{
	Value result;
	Module *pModule = new Module(&env, AScript_Symbol(_anonymous_));
	result.InitAsModule(pModule);
	context.GetBlock()->Exec(*pModule, sig);
	return result;
}

// mixin(module:Module)
AScript_DeclareFunction(mixin)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "module", VTYPE_Module);
}

AScript_ImplementFunction(mixin)
{
	
	return Value::Null;
}

// import(`module+) {symbols?}
AScript_DeclareFunctionEx(import_, "import")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "module", VTYPE_Quote, false, OCCUR_OnceOrMore);
	//DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(import_)
{
	SymbolSet symbolSet;
	if (context.IsBlockSpecified()) {
		foreach_const (ExprList, ppExpr, context.GetBlock()->GetExprList()) {
			if (!(*ppExpr)->IsSymbol()) {
				sig.SetError(ERR_SyntaxError,
					"wrong format for an element in import list");
				return Value::Null;
			}
			const Expr_Symbol *pExprSymbol =
							dynamic_cast<const Expr_Symbol *>(*ppExpr);
			symbolSet.insert(pExprSymbol->GetSymbol());
		}
	}
	foreach_const (ValueList, pValue, context.GetList(0)) {
		if (!env.ImportModule(sig, pValue->GetExpr(), symbolSet)) {
			return Value::Null;
		}
	}
	return Value::Null;
}

// execfile(filename:string):map
AScript_DeclareFunction(execfile)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "filename", VTYPE_String);
}

AScript_ImplementFunction(execfile)
{
	File file;
	file.Open(sig, context.GetString(0), "rt", NULL);
	if (!sig.IsSignalled()) {
		Parser().ExecFile(env, sig, file, true);
	}
	return Value::Null;
}

// expr = parse(code:string):map
AScript_DeclareFunction(parse)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "code", VTYPE_String);
}

AScript_ImplementFunction(parse)
{
	Value result;
	Expr *pExpr = Parser().ParseString(env, sig, context.GetString(0));
	if (pExpr == NULL) return Value::Null;
	result.InitAsExpr(env, pExpr);
	return result;
}

// val = eval(expr:Expr):map
AScript_DeclareFunction(eval)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "expr", VTYPE_Expr);
}

AScript_ImplementFunction(eval)
{
	return context.GetExpr(0)->Exec(env, sig);
}

// val = locals()
AScript_DeclareFunction(locals)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementFunction(locals)
{
	Value value;
	value.InitAsEnvironment(env);
	return value;
}

// val = outers()
AScript_DeclareFunction(outers)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementFunction(outers)
{
	Value value;
	Environment envOuter(&env, ENVTYPE_Outer);
	value.InitAsEnvironment(envOuter);
	return value;
}

// try () {block}
class Func_try : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	virtual bool IsSucceedable(const ICallable *pCallable) const;
	Func_try(Environment &env, const char *name = "try");
};
Func_try::Func_try(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareBlock(OCCUR_Once);
}

Value Func_try::DoEval(Environment &env, Signal sig, Context &context) const
{
	Value result = context.GetBlock()->Exec(env, sig);
	sig.SaveError();
	sig.ClearSignal();
	context.RequestSucceeding(this);
	return result;
}

bool Func_try::IsSucceedable(const ICallable *pCallable) const
{
	return true;
}

// except (error*:Error) {block}
class Func_except : public Function {
private:
	SymbolSet _symbolsSucceed;
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	virtual bool IsSucceedable(const ICallable *pCallable) const;
	Func_except(Environment &env, const char *name = "except");
};
Func_except::Func_except(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "error", VTYPE_Error, false, OCCUR_ZeroOrMore);
	DeclareBlock(OCCUR_Once);
	_symbolsSucceed.Insert(AScript_Symbol(except_));
}

Value Func_except::DoEval(Environment &env, Signal sig, Context &context) const
{
	if (sig.GetSavedErrorType() != ERR_None) {
		bool matchFlag = context.GetList(0).empty();
		foreach_const (ValueList, pValue, context.GetList(0)) {
			if (pValue->GetErrorType() == sig.GetSavedErrorType()) {
				matchFlag = true;
				break;
			}
		}
		if (matchFlag) {
			const Function *pFuncBlock = GetBlockFunction(env, sig, context);
			if (sig.IsSignalled()) return Value::Null;
			Environment envBlock(&env, ENVTYPE_Block);
			ValueList valListArg(Value(env, sig.GetSavedString()));
			Context contextSub(valListArg);
			return pFuncBlock->Eval(envBlock, sig, contextSub);
		}
	}
	context.RequestSucceeding(this);
	return Value::Null;
}

bool Func_except::IsSucceedable(const ICallable *pCallable) const
{
	return true;
}

// if (`cond) {block}
class Func_if : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	virtual bool IsSucceedable(const ICallable *pCallable) const;
	Func_if(Environment &env, const char *name = "if");
};
Func_if::Func_if(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "cond", VTYPE_Quote);
	DeclareBlock(OCCUR_Once);
}

Value Func_if::DoEval(Environment &env, Signal sig, Context &context) const
{
	Value value = context.GetExpr(0)->Exec(env, sig);
	if (value.GetBoolean()) {
		return context.GetBlock()->Exec(env, sig);
	}
	context.RequestSucceeding(this);
	return Value::Null;
}

bool Func_if::IsSucceedable(const ICallable *pCallable) const
{
	return true;
}

// elsif (`cond) {block}
class Func_elsif : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	virtual bool IsSucceedable(const ICallable *pCallable) const;
	Func_elsif(Environment &env, const char *name = "elsif");
};
Func_elsif::Func_elsif(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "cond", VTYPE_Quote);
	DeclareBlock(OCCUR_Once);
}

Value Func_elsif::DoEval(Environment &env, Signal sig, Context &context) const
{
	Value value = context.GetExpr(0)->Exec(env, sig);
	if (value.GetBoolean()) {
		return context.GetBlock()->Exec(env, sig);
	}
	context.RequestSucceeding(this);
	return Value::Null;
}

bool Func_elsif::IsSucceedable(const ICallable *pCallable) const
{
	return true;
}

// else () {block}
class Func_else : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_else(Environment &env, const char *name = "else");
};
Func_else::Func_else(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareBlock(OCCUR_Once);
}

Value Func_else::DoEval(Environment &env, Signal sig, Context &context) const
{
	return context.GetBlock()->Exec(env, sig);
}

// repeat (n?:number) {block}, option:list,xlist,set,xset
AScript_DeclareFunction(repeat)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "n", VTYPE_Number, false, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_Once);
}

AScript_ImplementFunction(repeat)
{
	Value result;
	const Value &value = context.GetValue(0);
	const Function *pFuncBlock = GetBlockFunction(env, sig, context);
	if (pFuncBlock == NULL) return Value::Null;
	if (context.IsRsltList() || context.IsRsltXList() ||
								context.IsRsltSet() || context.IsRsltXSet()) {
		if (!value.IsNumber()) {
			sig.SetError(ERR_AttributeError,
				"infinite loop is not allowed with list or xlist");
			return Value::Null;
		}
		bool removeInvalidFlag = context.IsRsltXList() || context.IsRsltXSet();
		bool listFlag = context.IsRsltList() || context.IsRsltXList();
		ValueList &valList = result.InitAsList(env);
		int cnt = static_cast<int>(value.GetNumber());
		for (int idx = 0; cnt > 0; cnt--, idx++) {
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			Value resultElem = pFuncBlock->Eval(env, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, resultElem, Value::Null)
			if (removeInvalidFlag && resultElem.IsInvalid()) {
				// nothing to do
			} else if (listFlag || !valList.IsContain(resultElem)) {
				valList.push_back(resultElem);
			}
		}
	} else {
		int cnt, cntStep;
		if (value.IsNumber()) {
			cnt = static_cast<int>(value.GetNumber());
			cntStep = -1;
		} else {
			cnt = 1;
			cntStep = 0;
		}
		for (int idx = 0; cnt > 0; cnt += cntStep, idx++) {
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			result = pFuncBlock->Eval(env, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, result, Value::Null)
		}
	}
	return result;
}

// while (`cond) {block}, option:list,xlist,set,xset
class Func_while : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_while(Environment &env, const char *name = "while");
};
Func_while::Func_while(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "cond", VTYPE_Quote);
	DeclareBlock(OCCUR_Once);
}

Value Func_while::DoEval(Environment &env, Signal sig, Context &context) const
{
	Value result;
	const Expr *pExpr = context.GetExpr(0);
	const Function *pFuncBlock = GetBlockFunction(env, sig, context);
	if (pFuncBlock == NULL) return Value::Null;
	if (context.IsRsltList() || context.IsRsltXList() ||
								context.IsRsltSet() || context.IsRsltXSet()) {
		bool removeInvalidFlag = context.IsRsltXList() || context.IsRsltXSet();
		bool listFlag = context.IsRsltList() || context.IsRsltXList();
		ValueList &valList = result.InitAsList(env);
		for (int idx = 0; pExpr->Exec(env, sig).GetBoolean(); idx++) {
			if (sig.IsSignalled()) return Value::Null;
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			Value resultElem = pFuncBlock->Eval(env, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, resultElem, Value::Null)
			if (removeInvalidFlag && resultElem.IsInvalid()) {
				// nothing to do
			} else if (listFlag || !valList.IsContain(resultElem)) {
				valList.push_back(resultElem);
			}
		}
	} else {
		for (int idx = 0; pExpr->Exec(env, sig).GetBoolean(); idx++) {
			if (sig.IsSignalled()) return Value::Null;
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			result = pFuncBlock->Eval(env, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, result, Value::Null)
		}
	}
	return result;
}

// for (`iter+) {block}, option:list,xlist,set,xset
class Func_for : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_for(Environment &env, const char *name = "for");
};
Func_for::Func_for(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "iter", VTYPE_Quote, false, OCCUR_OnceOrMore);
	DeclareBlock(OCCUR_Once);
}

Value Func_for::DoEval(Environment &env, Signal sig, Context &context) const
{
	SymbolList symbolList;
	IteratorOwner iteratorOwner;
	ValueList valListSave;	// save value to prevent them from being deleted
	Environment envBlock(&env, ENVTYPE_Block);
	const Function *pFuncBlock = GetBlockFunction(envBlock, sig, context);
	if (pFuncBlock == NULL) return Value::Null;
	foreach_const (ValueList, pValue, context.GetList(0)) {
		ASSUME(env, pValue->IsExpr());
		const Symbol *pSymbol;
		const Expr *pExpr = pValue->GetExpr();
		Value value;
		if (pExpr->IsContainCheck()) {
			const Expr_BinaryOp *pExprBin =
								dynamic_cast<const Expr_BinaryOp *>(pExpr);
			if (!pExprBin->GetLeft()->IsSymbol()) {
				sig.SetError(ERR_ValueError,
					"l-value of assignment in foreach() arguments must be a symbol");
				return Value::Null;
			}
			const Expr_Symbol *pExprSym =
					dynamic_cast<const Expr_Symbol *>(pExprBin->GetLeft());
			pSymbol = pExprSym->GetSymbol();
			value = pExprBin->GetRight()->Exec(envBlock, sig);
		} else if (pExpr->IsSymbol()) {
			const Expr_Symbol *pExprSym = dynamic_cast<const Expr_Symbol *>(pExpr);
			pSymbol = pExprSym->GetSymbol();
			value = pExpr->Exec(envBlock, sig);
		} else {
			sig.SetError(ERR_TypeError,
				"for() can take iterator or symbol as its argument");
			return Value::Null;
		}
		valListSave.push_back(value);
		if (sig.IsSignalled()) return Value::Null;
		Iterator *pIterator = value.CreateIterator(sig);
		if (pIterator == NULL) return Value::Null;
		symbolList.push_back(pSymbol);
		iteratorOwner.push_back(pIterator);
	}
	Value result;
	if (context.IsRsltList() || context.IsRsltXList() ||
								context.IsRsltSet() || context.IsRsltXSet()) {
		bool removeInvalidFlag = context.IsRsltXList() || context.IsRsltXSet();
		bool listFlag = context.IsRsltList() || context.IsRsltXList();
		ValueList &valList = result.InitAsList(env);
		for (int idx = 0; ; idx++) {
			SymbolList::iterator ppSymbol = symbolList.begin();
			bool doneFlag = false;
			foreach_const (IteratorOwner, ppIterator, iteratorOwner) {
				Iterator *pIterator = *ppIterator;
				Value value;
				if (!pIterator->Next(sig, value)) {
					if (sig.IsSignalled()) return Value::Null;
					doneFlag = true;
					break;
				}
				envBlock.AssignValue(*ppSymbol, value, false);
				ppSymbol++;
			}
			if (doneFlag) break;
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			Value resultElem = pFuncBlock->Eval(envBlock, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, resultElem, Value::Null)
			if (removeInvalidFlag && resultElem.IsInvalid()) {
				// nothing to do
			} else if (listFlag || !valList.IsContain(resultElem)) {
				valList.push_back(resultElem);
			}
		}
	} else {
		//for (int idx = 0; iteratorOwner.IsValidAll(); iteratorOwner.Next(), idx++) {
		for (int idx = 0; ; idx++) {
			SymbolList::iterator ppSymbol = symbolList.begin();
			bool doneFlag = false;
			foreach_const (IteratorOwner, ppIterator, iteratorOwner) {
				Iterator *pIterator = *ppIterator;
				Value value;
				if (!pIterator->Next(sig, value)) {
					if (sig.IsSignalled()) return Value::Null;
					doneFlag = true;
					break;
				}
				envBlock.AssignValue(*ppSymbol, value, false);
				ppSymbol++;
			}
			if (doneFlag) break;
			ValueList valListArg(Value(static_cast<Number>(idx)));
			Context contextSub(valListArg);
			result = pFuncBlock->Eval(envBlock, sig, contextSub);
			AScript_BlockSignalHandlerInLoop(sig, result, Value::Null)
		}
	}
	return result;
}

// break(value?)
class Func_break : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_break(Environment &env, const char *name = "break");
};
Func_break::Func_break(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
}

Value Func_break::DoEval(Environment &env, Signal sig, Context &context) const
{
	sig.SetSignal(SIGTYPE_Break, context.GetValue(0));
	return Value::Null;
}

// continue(value?)
class Func_continue : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_continue(Environment &env, const char *name = "continue");
};
Func_continue::Func_continue(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
}

Value Func_continue::DoEval(Environment &env, Signal sig, Context &context) const
{
	sig.SetSignal(SIGTYPE_Continue, context.GetValue(0));
	return Value::Null;
}

// return(value?)
class Func_return : public Function {
public:
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
	Func_return(Environment &env, const char *name = "return");
};
Func_return::Func_return(Environment &env, const char *name) : Function(name, FUNCTYPE_Function)
{
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
}

Value Func_return::DoEval(Environment &env, Signal sig, Context &context) const
{
	sig.SetSignal(SIGTYPE_Return, context.GetValue(0));
	return Value::Null;
}

// raise(error, msg?:string)
AScript_DeclareFunction(raise)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "error", VTYPE_Error);
	DeclareArg(env, "msg", VTYPE_String, false, OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(raise)
{
	sig.SetError(context.GetErrorType(0), "%s",
					context.IsString(1)? context.GetString(1) : "error");
	return Value::Null;
}

// list = zip(value+):map
AScript_DeclareFunction(zip)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_OnceOrMore);
}

AScript_ImplementFunction(zip)
{
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		valList.push_back(*pValue);
	}
	return result;
}

// list = sort(elem[], directive?):[stable]
AScript_DeclareFunction(sort)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "directive", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(stable));
}

AScript_ImplementFunction(sort)
{
	Object_List *pObj = context.GetListObj(0);
	return pObj->SortRank(sig, context.GetValue(1),
							false, context.IsSet(AScript_Symbol(stable)));
}

// list = rank(elem[], directive?):[stable]
AScript_DeclareFunction(rank)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "directive", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(stable));
}

AScript_ImplementFunction(rank)
{
	Object_List *pObj = context.GetListObj(0);
	return pObj->SortRank(sig, context.GetValue(1),
							true, context.IsSet(AScript_Symbol(stable)));
}

// iter = reverse(elem[])
AScript_DeclareFunction(reverse)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(reverse)
{
	Object_List *pObj = dynamic_cast<Object_List *>(context.GetListObj(0)->IncRef());
	return ReturnIterator(env, sig, context,
							new Object_List::IteratorReverse(pObj));
}

// iter = round(elem[], n?:number)
AScript_DeclareFunction(round)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "n", VTYPE_Number, false, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(round)
{
	int cnt = context.IsValid(1)? static_cast<int>(context.GetNumber(1)) : -1;
	Object_List *pObj = dynamic_cast<Object_List *>(context.GetListObj(0)->IncRef());
	return ReturnIterator(env, sig, context,
							new Object_List::IteratorRound(pObj, cnt));
}

// num = len(iter:iterator)
AScript_DeclareFunction(len)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator);
}

AScript_ImplementFunction(len)
{
	int cnt = 0;
	Value value;
	for (Iterator *pIterator = context.GetIterator(0);
										pIterator->Next(sig, value); cnt++) ;
	return Value(static_cast<Number>(cnt));
}

// list = list(iter+:iterator)
AScript_DeclareFunction(list)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator, false, OCCUR_OnceOrMore);
}

AScript_ImplementFunction(list)
{
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		Value value;
		for (Iterator *pIterator = pValue->GetIterator(); pIterator->Next(sig, value); ) {
			valList.push_back(value);
		}
		if (sig.IsSignalled()) return Value::Null;
	}
	return result;
}

// list = xlist(iter+:iterator)
AScript_DeclareFunction(xlist)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator, false, OCCUR_OnceOrMore);
}

AScript_ImplementFunction(xlist)
{
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		Value value;
		for (Iterator *pIterator = pValue->GetIterator(); pIterator->Next(sig, value); ) {
			if (value.IsValid()) valList.push_back(value);
		}
		if (sig.IsSignalled()) return Value::Null;
	}
	return result;
}

// list = set(iter+:iterator)
// list = xset(iter+:iterator)
// attrs: and, xor, or
AScript_DeclareFunctionBegin(set_xset)
	bool _acceptInvalidFlag;
AScript_DeclareFunctionEnd(set_xset)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	_acceptInvalidFlag = (::strcmp(GetName(), "set") == 0);
	DeclareArg(env, "iter", VTYPE_Iterator, false, OCCUR_OnceOrMore);
	DeclareAttr(AScript_Symbol(or));
	DeclareAttr(AScript_Symbol(and));
	DeclareAttr(AScript_Symbol(xor));
}

AScript_ImplementFunction(set_xset)
{
	Value result;
	ValueList &valList = result.InitAsList(env);
	if (context.IsSet(AScript_Symbol(and))) {			// AND combination
		ValueList valList1, valList2;
		ValueList::const_reverse_iterator pValueArg = context.GetList(0).rbegin();
		Value value;
		for (Iterator *pIterator = pValueArg->GetIterator();
											pIterator->Next(sig, value); ) {
			if ((_acceptInvalidFlag || value.IsValid()) &&
											!valList1.IsContain(value)) {
				valList1.push_back(value);
			}
		}
		if (sig.IsSignalled()) return Value::Null;
		pValueArg++;
		ValueList *pValListAnd = &valList1;
		ValueList *pValListWork = &valList2;
		for ( ; pValueArg != context.GetList(0).rend() &&
										!pValListAnd->empty(); pValueArg++) {
			Value value;
			for (Iterator *pIterator = pValueArg->GetIterator();
											pIterator->Next(sig, value); ) {
				if (pValListAnd->IsContain(value) && !pValListWork->IsContain(value)) {
					pValListWork->push_back(value);
				}
			}
			ValueList *pValListTmp = pValListAnd;
			pValListAnd = pValListWork, pValListWork = pValListTmp;
			pValListWork->clear();
		}
		foreach_const (ValueList, pValue, *pValListAnd) {
			valList.push_back(*pValue);
		}
	} else if (context.IsSet(AScript_Symbol(xor))) {	// XOR combination
		ValueList valList1, valList2;
		ValueList *pValListAnd = &valList1;
		ValueList valListOr;
		ValueList::const_iterator pValueArg = context.GetList(0).begin();
		Value value;
		for (Iterator *pIterator = pValueArg->GetIterator();
										pIterator->Next(sig, value); ) {
			if (!valList1.IsContain(value)) valList1.push_back(value);
			if ((_acceptInvalidFlag || value.IsValid()) &&
											!valListOr.IsContain(value)) {
				valListOr.push_back(value);
			}
		}
		if (sig.IsSignalled()) return Value::Null;
		pValueArg++;
		ValueList *pValListWork = &valList2;
		for ( ; pValueArg != context.GetList(0).end() &&
										!pValListAnd->empty(); pValueArg++) {
			Value value;
			for (Iterator *pIterator = pValueArg->GetIterator();
										pIterator->Next(sig, value); ) {
				if (pValListAnd->IsContain(value)) pValListWork->push_back(value);
				if ((_acceptInvalidFlag || value.IsValid()) &&
												!valListOr.IsContain(value)) {
					valListOr.push_back(value);
				}
			}
			if (sig.IsSignalled()) return Value::Null;
			ValueList *pValListTmp = pValListAnd;
			pValListAnd = pValListWork, pValListWork = pValListTmp;
			pValListWork->clear();
		}
		foreach_const (ValueList, pValue, valListOr) {
			if (!pValListAnd->IsContain(*pValue)) {
				valList.push_back(*pValue);
			}
		}
	} else {										// OR combination
		foreach_const (ValueList, pValue, context.GetList(0)) {
			Value value;
			for (Iterator *pIterator = pValue->GetIterator();
												pIterator->Next(sig, value); ) {
				if ((_acceptInvalidFlag || value.IsValid()) &&
												!valList.IsContain(value)) {
					valList.push_back(value);
				}
			}
			if (sig.IsSignalled()) return Value::Null;
		}
	}
	return result;
}

// replace(elem[], value, replace)
AScript_DeclareFunction(replace)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "value", VTYPE_AnyType);
	DeclareArg(env, "replace", VTYPE_AnyType);
}

AScript_ImplementFunction(replace)
{
	const Value &value = context.GetValue(1);
	const Value &replace = context.GetValue(2);
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		if (Value::Compare(*pValue, value) == 0) {
			valList.push_back(replace);
		} else {
			valList.push_back(*pValue);
		}
	}
	return result;
}

// nil_to(elem[], replace)
AScript_DeclareFunction(nil_to)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "replace", VTYPE_AnyType);
}

AScript_ImplementFunction(nil_to)
{
	const Value &replace = context.GetValue(1);
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, context.GetList(0)) {
		if (pValue->IsValid()) {
			valList.push_back(*pValue);
		} else {
			valList.push_back(replace);
		}
	}
	return result;
}

// filter(criteria:Function, iter:iterator) {block?}
AScript_DeclareFunction(filter)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "criteria", VTYPE_Function);
	DeclareArg(env, "iter", VTYPE_Iterator);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(filter)
{
	Object_Iterator *pSelf = Object_Iterator::GetSelfObj(context);
	Iterator *pIterator = new Iterator_FilterByFunction(env,
			context.GetIterator(1)->IncRef(),
			dynamic_cast<Object_Function *>(context.GetFunctionObj(0)->IncRef()));
	//Iterator *pIterator =
	//			context.GetIterator(1)->Filter(env, sig, context.GetValue(0));
	return ReturnIterator(env, sig, context, pIterator);
}

// list = map(func:Function, elem[])
AScript_DeclareFunction(map)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "func", VTYPE_Function);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
}

AScript_ImplementFunction(map)
{
	const Value &value1 = context.GetValue(0);
	const Value &value2 = context.GetValue(1);
	const Function *pFunc = value1.GetFunction();
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueList, pValue, value2.GetList()) {
		ValueList valListArg(*pValue);
		Context contextSub(context, valListArg, Value::Null);
		Value value = pFunc->Eval(env, sig, contextSub);
		if (sig.IsSignalled()) return Value::Null;
		valList.push_back(value);
	}
	return result;
}

// value = reduce(func:Function, elem[], value0)
AScript_DeclareFunction(reduce)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "func", VTYPE_Function);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
	DeclareArg(env, "value0", VTYPE_AnyType);
}

AScript_ImplementFunction(reduce)
{
	const Value &value1 = context.GetValue(0);
	const Value &value2 = context.GetValue(1);
	const Function *pFunc = value1.GetFunction();
	Value result = context.GetValue(2);
	foreach_const (ValueList, pValue, value2.GetList()) {
		ValueList valListArg(result, *pValue);
		Context contextSub(context, valListArg, Value::Null);
		result = pFunc->Eval(env, sig, contextSub);
		if (sig.IsSignalled()) return Value::Null;
	}
	return result;
}

// num = last_index(elem[]):map
AScript_DeclareFunction(last_index)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "list", VTYPE_AnyType, true);
}

AScript_ImplementFunction(last_index)
{
	return Value(static_cast<Number>(context.GetList(0).size()) - 1);
}

// v = min(iter:iterator):[index,last_index,indices]
AScript_DeclareFunction(min)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator);
	DeclareAttr(AScript_Symbol(index));
	DeclareAttr(AScript_Symbol(last_index));
	DeclareAttr(AScript_Symbol(indices));
}

AScript_ImplementFunction(min)
{
	Iterator *pIterator = context.GetIterator(0);
	return pIterator->MinMax(env, sig, false, context.GetAttrs());
}

// v = max(iter:iterator):[index,last_index,indices]
AScript_DeclareFunction(max)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator);
	DeclareAttr(AScript_Symbol(index));
	DeclareAttr(AScript_Symbol(last_index));
	DeclareAttr(AScript_Symbol(indices));
}

AScript_ImplementFunction(max)
{
	Iterator *pIterator = context.GetIterator(0);
	return pIterator->MinMax(env, sig, true, context.GetAttrs());
}

// v = choosemin(values+):map / v = choosemax(values+):map
AScript_DeclareFunctionBegin(choosemin_max)
	bool _maxFlag;
AScript_DeclareFunctionEnd(choosemin_max)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	_maxFlag = (::strcmp(GetName(), "choosemax") == 0);
	DeclareArg(env, "values", VTYPE_AnyType, false, OCCUR_OnceOrMore);
}

AScript_ImplementFunction(choosemin_max)
{
	const ValueList &valList = context.GetList(0);
	ValueList::const_iterator pValue = valList.begin();
	Value result = *pValue++;
	for ( ; pValue != valList.end(); pValue++) {
		int cmp = Value::Compare(result, *pValue);
		if (_maxFlag) cmp = -cmp;
		if (cmp > 0) result = *pValue;
	}
	return result;
}

// v = choose(index:number, values+):map
AScript_DeclareFunction(choose)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "index", VTYPE_Number);
	DeclareArg(env, "values", VTYPE_AnyType, false, OCCUR_OnceOrMore);
}

AScript_ImplementFunction(choose)
{
	size_t index = static_cast<size_t>(context.GetNumber(0));
	const ValueList &valList = context.GetList(1);
	if (index >= valList.size()) {
		sig.SetError(ERR_IndexError, "index is out of range");
		return Value::Null;
	}
	return valList[index];
}

// v = chooseif(flag, value1, vlaue2):map
AScript_DeclareFunction(chooseif)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "flag", VTYPE_Boolean);
	DeclareArg(env, "value1", VTYPE_AnyType);
	DeclareArg(env, "value2", VTYPE_AnyType);
}

AScript_ImplementFunction(chooseif)
{
	return context.GetBoolean(0)? context.GetValue(1) : context.GetValue(2);
}

// num = sum(iter:iterator)
AScript_DeclareFunction(sum)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator);
}

AScript_ImplementFunction(sum)
{
	Iterator *pIterator = context.GetIterator(0);
	Value value;
	if (!pIterator->Next(sig, value)) return Value::Null;
	Value result = value;
	while (pIterator->Next(sig, value)) {
		ValueList valListArg(result, value);
		Context contextSub(valListArg);
		result = env.GetFunc_Plus().Eval(env, sig, contextSub);
		if (sig.IsSignalled()) return Value::Null;
	}
	if (sig.IsSignalled()) return Value::Null;
	return result;
}

// num = average(iter:iterator)
AScript_DeclareFunction(average)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "iter", VTYPE_Iterator);
}

AScript_ImplementFunction(average)
{
	int cnt = 0;
	Iterator *pIterator = context.GetIterator(0);
	Value value;
	if (!pIterator->Next(sig, value)) return Value::Null;
	Value result = value;
	for (cnt = 1; pIterator->Next(sig, value); cnt++) {
		ValueList valListArg(result, value);
		Context contextSub(valListArg);
		result = env.GetFunc_Plus().Eval(env, sig, contextSub);
		if (sig.IsSignalled()) return Value::Null;
	}
	if (sig.IsSignalled()) return Value::Null;
	ValueList valListArg(result, Value(static_cast<Number>(cnt)));
	Context contextSub(valListArg);
	return env.GetFunc_Divide().Eval(env, sig, contextSub);
}

// flag = and(elem[]:boolean):map
AScript_DeclareFunctionEx(and_, "and")
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_Boolean, true);
}

AScript_ImplementFunction(and_)
{
	const ValueList &valList = context.GetList(0);
	if (valList.empty()) return Value::Null;
	foreach_const (ValueList, pValue, valList) {
		if (!pValue->GetBoolean()) return Value(false);
	}
	return Value(true);
}

// flag = or(elem[]:boolean):map
AScript_DeclareFunctionEx(or_, "or")
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_Boolean, true);
}

AScript_ImplementFunction(or_)
{
	const ValueList &valList = context.GetList(0);
	if (valList.empty()) return Value::Null;
	foreach_const (ValueList, pValue, valList) {
		if (pValue->GetBoolean()) return Value(true);
	}
	return Value(false);
}

// text = to_string(value):map
AScript_DeclareFunction(to_string)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType);
}

AScript_ImplementFunction(to_string)
{
	return Value(env, context.GetValue(0).ToString(sig, false).c_str());
}

// num = to_number(value):map:[strict,raise,zero,nil]
AScript_DeclareFunction(to_number)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType);
	DeclareAttr(AScript_Symbol(strict));
	DeclareAttr(AScript_Symbol(raise));
	DeclareAttr(AScript_Symbol(zero));
	DeclareAttr(AScript_Symbol(nil));
}

AScript_ImplementFunction(to_number)
{
	bool allowPartFlag = !context.IsSet(AScript_Symbol(strict));
	bool successFlag;
	Number num = context.GetValue(0).ToNumber(allowPartFlag, successFlag);
	if (successFlag) {
		return Value(num);
	} else if (context.IsSet(AScript_Symbol(raise))) {
		sig.SetError(ERR_ValueError, "failed to convert to a number");
		return Value::Null;
	} else if (context.IsSet(AScript_Symbol(zero))) {
		return Value(0.);
	} else { // context.IsSet(AScript_PrivSymbol(nil)
		return Value::Null;
	}
}

// symbol = to_symbol(str:string):map
AScript_DeclareFunction(to_symbol)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "str", VTYPE_String);
}

AScript_ImplementFunction(to_symbol)
{
	return Value(Symbol::Add(context.GetString(0)));
}

// num = rand(num?)
AScript_DeclareFunction(rand)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "num", VTYPE_Number, false, OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(rand)
{
	if (context.IsInvalid(0)) {
		return Value(::genrand_real2());
	}
	unsigned long num = static_cast<unsigned long>(context.GetNumber(0));
	Number result = static_cast<unsigned long>(::genrand_real2() * num);
	return Value(result);
}

// ord(str:string):map
AScript_DeclareFunction(ord)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "str", VTYPE_String);
}

AScript_ImplementFunction(ord)
{
	const char *str = context.GetString(0);
	unsigned long num = static_cast<unsigned char>(*str);
	if (IsUTF8First(*str)) {
		str++;
		for ( ; IsUTF8Follower(*str); str++) {
			num = (num << 8) | static_cast<unsigned char>(*str);
		}
	}
	return Value(static_cast<Number>(num));
}

// chr(num:number):map
AScript_DeclareFunction(chr)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "num", VTYPE_Number);
}

AScript_ImplementFunction(chr)
{
	unsigned long num = static_cast<unsigned long>(context.GetNumber(0));
	int i = 0;
	char buff[4];
	buff[i++] = static_cast<unsigned char>(num & 0xff);
	num >>= 8;
	for ( ; num != 0; num >>= 8) {
		buff[i++] = static_cast<unsigned char>(num & 0xff);
	}
	String str;
	for ( ; i > 0; i--) {
		str.push_back(buff[i - 1]);
	}
	return Value(env, str.c_str());
}

// print(value*):map:void
AScript_DeclareFunction(print)
{
	SetMode(RSLTMODE_Void, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_ZeroOrMore);
}

AScript_ImplementFunction(print)
{
	foreach_const (ValueList, pValue, context.GetList(0)) {
		env.PutValue(sig, *pValue);
	}
	return Value::Null;
}

// println(value*):map:void
AScript_DeclareFunction(println)
{
	SetMode(RSLTMODE_Void, MAP_On, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType, false, OCCUR_ZeroOrMore);
}

AScript_ImplementFunction(println)
{
	foreach_const (ValueList, pValue, context.GetList(0)) {
		env.PutValue(sig, *pValue);
	}
	env.PutChar('\n');
	return Value::Null;
}

// printf(format, values*):map:void
AScript_DeclareFunction(printf)
{
	SetMode(RSLTMODE_Void, MAP_On, FLAT_Off);
	DeclareArg(env, "format", VTYPE_String);
	DeclareArg(env, "values", VTYPE_AnyType, false, OCCUR_ZeroOrMore);
}

AScript_ImplementFunction(printf)
{
	Value value(env, Formatter::Format(sig,
							context.GetString(0), context.GetList(1)).c_str());
	env.PutValue(sig, value);
	return Value::Null;
}

// str = sprintf(format, values*):map
AScript_DeclareFunction(sprintf)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "format", VTYPE_String);
	DeclareArg(env, "values", VTYPE_AnyType, false, OCCUR_ZeroOrMore);
}

AScript_ImplementFunction(sprintf)
{
	return Value(env, Formatter::Format(sig,
						context.GetString(0), context.GetList(1)).c_str());
}

// dir(obj?)
AScript_DeclareFunction(dir)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "obj", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(dir)
{
	const Environment::Frame *pFrame = &env.GetBottomFrame();
	const Value &value = context.GetValue(0);
	if (value.IsModule()) {
		pFrame = &value.GetModule()->GetTopFrame();
	} else if (value.IsFunction() &&
						value.GetFunction()->GetClassToConstruct() != NULL) {
		const Class *pClass = value.GetFunction()->GetClassToConstruct();
		pFrame = &pClass->GetTopFrame();
	} else if (value.IsObject()) {
		const Object *pObj = value.GetObject();
		Value result;
		ValueList &valList = result.InitAsList(env);
		foreach_const (Environment::FrameList, ppFrame, pObj->GetFrameList()) {
			const Environment::Frame *pFrame = *ppFrame;
			if (!pFrame->IsType(ENVTYPE_Class) && !pFrame->IsType(ENVTYPE_Instance)) {
				break;
			}
			foreach_const (ValueMap, iter, pFrame->GetValueMap()) {
				valList.push_back(Value(env, iter->first->GetName()));
			}
		}
		return result;
	} else if (value.IsValid()) {
		sig.SetError(ERR_TypeError, "invalid argument");
		return Value::Null;
	}
	Value result;
	ValueList &valList = result.InitAsList(env);
	foreach_const (ValueMap, iter, pFrame->GetValueMap()) {
		valList.push_back(Value(env, iter->first->GetName()));
	}
	return result;
}

// is_defined(`symbol)
AScript_DeclareFunction(is_defined)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "symbol", VTYPE_Quote);
}

AScript_ImplementFunction(is_defined)
{
	bool definedFlag = false;
	const Expr *pExpr = context.GetExpr(0);
	if (pExpr->IsSymbol() || pExpr->IsField()) {
		pExpr->Exec(env, sig);
		if (sig.IsSignalled() && !sig.IsError()) return Value::Null;
		definedFlag = !sig.IsError();
		sig.ClearSignal();
	} else {
		sig.SetError(ERR_ValueError, "argument must be a symbol");
		return Value::Null;
	}
	return Value(definedFlag);
}

// is_type*(value)
class AScript_Function(is_type_) : public Function {
private:
	ValueType _valType;
public:
	AScript_Function(is_type_)(Environment &env, const char *name, ValueType valType);
	virtual Value DoEval(Environment &env, Signal sig, Context &context) const;
};
AScript_Function(is_type_)::AScript_Function(is_type_)(
					Environment &env, const char *name, ValueType valType) :
	Function(name, FUNCTYPE_Function), _valType(valType)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "value", VTYPE_AnyType);
}

AScript_ImplementFunction(is_type_)
{
	return Value(context.GetValue(0).GetType() == _valType);
}

// typename(`value)
AScript_DeclareFunctionEx(typename_, "typename")
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "value", VTYPE_Quote);
}

AScript_ImplementFunction(typename_)
{
	const Expr *pExpr = context.GetExpr(0);
	const char *typeName = "unknown";
	if (pExpr->IsSymbol() || pExpr->IsField()) {
		Value value = pExpr->Exec(env, sig);
		if (sig.IsSignalled() && !sig.IsError()) return Value::Null;
		typeName = sig.IsError()? "undefined" : value.GetTypeName();
		sig.ClearSignal();
	} else if (pExpr->IsValue()) {
		Value value = dynamic_cast<const Expr_Value *>(pExpr)->GetValue();
		typeName = value.GetTypeName();
	} else {
		sig.SetError(ERR_ValueError, "argument must be a symbol or value");
		return Value::Null;
	}
	return Value(env, typeName);
}

// iter = range(num:number, num_end?:number, step?:number):map
class Iterator_range : public Iterator {
private:
	Number _num;
	Number _numEnd;
	Number _numStep;
public:
	Iterator_range(Number numBegin, Number numEnd, Number numStep);
	~Iterator_range();
	virtual bool Next(Signal sig, Value &value);
};

Iterator_range::Iterator_range(Number numBegin, Number numEnd, Number numStep) :
		Iterator(false), _num(numBegin), _numEnd(numEnd), _numStep(numStep)
{
}

Iterator_range::~Iterator_range()
{
}

bool Iterator_range::Next(Signal sig, Value &value)
{
	if (!((_numStep > 0)? (_num < _numEnd) : (_num > _numEnd))) return false;
	value = Value(_num);
	_num += _numStep;
	return true;
}

AScript_DeclareFunction(range)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "num",		VTYPE_Number);
	DeclareArg(env, "num_end",	VTYPE_Number, false, OCCUR_ZeroOrOnce);
	DeclareArg(env, "step",		VTYPE_Number, false, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(range)
{
	Number numBegin = 0.;
	Number numEnd = 0.;
	Number numStep = 1.;
	if (context.GetValue(1).IsInvalid()) {
		numEnd = context.GetValue(0).GetNumber();
		if (numBegin > numEnd) numStep = -1.;
	} else if (context.GetValue(2).IsInvalid()) {
		numBegin = context.GetValue(0).GetNumber();
		numEnd = context.GetValue(1).GetNumber();
		if (numBegin > numEnd) numStep = -1.;
	} else {
		numBegin = context.GetValue(0).GetNumber();
		numEnd = context.GetValue(1).GetNumber();
		numStep = context.GetValue(2).GetNumber();
		if (numStep == 0.) {
			sig.SetError(ERR_ValueError, "step cannot be specified as zero");
			return Value::Null;
		}
	}
	//return Value(env, new Iterator_range(numBegin, numEnd, numStep));
	return ReturnIterator(env, sig, context,
						new Iterator_range(numBegin, numEnd, numStep));
}

// iter = interval(a:number, b:number, samples:number):map
class Iterator_interval : public Iterator {
private:
	Number _numBegin;
	Number _numEnd;
	Number _numDenom;
	int _numSamples;
	int _iFactor;
	int _idx;
public:
	Iterator_interval(Number numBegin, Number numEnd,
								Number numDenom, int numSamples, int iFactor);
	~Iterator_interval();
	virtual bool Next(Signal sig, Value &value);
};

Iterator_interval::Iterator_interval(Number numBegin, Number numEnd,
								Number numDenom, int numSamples, int iFactor) :
		Iterator(false), _numBegin(numBegin), _numEnd(numEnd), _numDenom(numDenom),
		_numSamples(numSamples), _iFactor(iFactor), _idx(0)
{
}

Iterator_interval::~Iterator_interval()
{
}

bool Iterator_interval::Next(Signal sig, Value &value)
{
	if (_idx >= _numSamples) return false;
	value = Value((_numEnd - _numBegin) * _iFactor / _numDenom + _numBegin);
	_iFactor++, _idx++;
	return true;
}

AScript_DeclareFunction(interval)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "a", VTYPE_Number);
	DeclareArg(env, "b", VTYPE_Number);
	DeclareArg(env, "samples", VTYPE_Number);
	DeclareAttr(AScript_PrivSymbol(open));
	DeclareAttr(AScript_PrivSymbol(open_l));
	DeclareAttr(AScript_PrivSymbol(open_r));
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementFunction(interval)
{
	const ValueList &valListArg = context.GetArgs();
	Number numBegin = valListArg[0].GetNumber();
	Number numEnd = valListArg[1].GetNumber();
	int numSamples = static_cast<int>(valListArg[2].GetNumber());
	if (numSamples <= 1) {
		sig.SetError(ERR_ValueError, "samples must be more than one");
		return Value::Null;
	}
	bool openFlag = context.IsSet(AScript_PrivSymbol(open));
	bool openLeftFlag = context.IsSet(AScript_PrivSymbol(open_l));
	bool openRightFlag = context.IsSet(AScript_PrivSymbol(open_r));
	int iFactor = 0;
	Number numDenom = numSamples - 1;
	if (openFlag || (openLeftFlag && openRightFlag)) {
		numDenom = numSamples + 1;
		iFactor = 1;
	} else if (openLeftFlag) {
		numDenom = numSamples;
		iFactor = 1;
	} else if (openRightFlag) {
		numDenom = numSamples;
		iFactor = 0;
	}
	return ReturnIterator(env, sig, context,
		new Iterator_interval(numBegin, numEnd, numDenom, numSamples, iFactor));
}

// Module entry
AScript_ModuleEntry()
{
	AScript_RealizePrivSymbol(open);
	AScript_RealizePrivSymbol(open_l);
	AScript_RealizePrivSymbol(open_r);
	AScript_AssignFunction(class_);
	env.AssignFunction(new Func_struct(env));
	AScript_AssignFunction(int_);
	env.AssignFunction(new Func_try(env));
	env.AssignFunction(new Func_except(env));
	env.AssignFunction(new Func_if(env));
	env.AssignFunction(new Func_elsif(env));
	env.AssignFunction(new Func_else(env));
	AScript_AssignFunction(repeat);
	env.AssignFunction(new Func_while(env));
	env.AssignFunction(new Func_for(env));
	env.AssignFunction(new Func_break(env));
	env.AssignFunction(new Func_continue(env));
	env.AssignFunction(new Func_return(env));
	AScript_AssignFunction(raise);
	AScript_AssignFunction(len);
	AScript_AssignFunction(list);
	AScript_AssignFunction(xlist);
	AScript_AssignFunctionEx(set_xset, "set");
	AScript_AssignFunctionEx(set_xset, "xset");
	AScript_AssignFunction(replace);
	AScript_AssignFunction(nil_to);
	AScript_AssignFunction(filter);
	AScript_AssignFunction(map);
	AScript_AssignFunction(reduce);
	AScript_AssignFunction(zip);
	AScript_AssignFunction(last_index);
	AScript_AssignFunction(sort);
	AScript_AssignFunction(rank);
	AScript_AssignFunction(reverse);
	AScript_AssignFunction(round);
	AScript_AssignFunction(min);
	AScript_AssignFunction(max);
	AScript_AssignFunctionEx(choosemin_max, "choosemin");
	AScript_AssignFunctionEx(choosemin_max, "choosemax");
	AScript_AssignFunction(choose);
	AScript_AssignFunction(chooseif);
	AScript_AssignFunction(sum);
	AScript_AssignFunction(average);
	AScript_AssignFunctionEx(and_, "and");
	AScript_AssignFunctionEx(or_, "or");
	AScript_AssignFunction(module);
	AScript_AssignFunction(mixin);
	AScript_AssignFunction(import_);
	AScript_AssignFunction(execfile);
	AScript_AssignFunction(parse);
	AScript_AssignFunction(eval);
	AScript_AssignFunction(locals);
	AScript_AssignFunction(outers);
	AScript_AssignFunction(to_string);
	AScript_AssignFunction(to_number);
	AScript_AssignFunction(to_symbol);
	AScript_AssignFunction(rand);
	AScript_AssignFunction(ord);
	AScript_AssignFunction(chr);
	AScript_AssignFunction(print);
	AScript_AssignFunction(println);
	AScript_AssignFunction(printf);
	AScript_AssignFunction(sprintf);
	AScript_AssignFunction(dir);
	AScript_AssignFunction(is_defined);
	AScript_AssignFunction(typename_);
	AScript_AssignFunctionExx(is_type_, "is_symbol", VTYPE_Symbol);
	AScript_AssignFunctionExx(is_type_, "is_boolean", VTYPE_Boolean);
	AScript_AssignFunctionExx(is_type_, "is_number", VTYPE_Number);
	AScript_AssignFunctionExx(is_type_, "is_complex", VTYPE_Complex);
	AScript_AssignFunctionExx(is_type_, "is_string", VTYPE_String);
	AScript_AssignFunctionExx(is_type_, "is_module", VTYPE_Module);
	AScript_AssignFunctionExx(is_type_, "is_class", VTYPE_Class);
	//AScript_AssignFunction(xrange);
	AScript_AssignFunction(range);
	AScript_AssignFunction(interval);
}

AScript_ModuleTerminate()
{
}

AScript_EndModule(__builtins__)

AScript_DLLModuleEntry(__builtins__)
