//
// Object_Dict
//
#include "Object_Dict.h"
#include "Expr.h"

namespace AScript {

//-----------------------------------------------------------------------------
// Object_Dict
//-----------------------------------------------------------------------------
bool Object_Dict::IsDict() const { return true; }

Object *Object_Dict::Clone() const
{
	return new Object_Dict(*this);
}

Value Object_Dict::GetByIndex(Signal sig, const Value &valueIdx)
{
	const Value *pValue = Find(sig, valueIdx);
	if (sig.IsSignalled()) {
		return Value::Null;
	} else if (pValue == NULL) {
		SetError_KeyNotFound(sig, valueIdx);
		return Value::Null;
	}
	return *pValue;
}

void Object_Dict::SetByIndex(Signal sig, const Value &valueIdx, const Value &value)
{
	if (!ValueDict::IsValidKey(valueIdx)) {
		sig.SetError(ERR_KeyError, "invalid value type for key");
		return;
	}
	_valDict[valueIdx] = value;
}

Iterator *Object_Dict::CreateIterator(Signal sig)
{
	return new IteratorPairs(dynamic_cast<Object_Dict *>(IncRef()));
}

String Object_Dict::ToString(Signal sig, bool exprFlag)
{
	String str;
	str += "%{";
	foreach_const (ValueDict, iter, _valDict) {
		if (iter != _valDict.begin()) str += ", ";
		const Value &value = iter->first;
		if (value.IsString()) {
			str += '"';
			str += value.GetString();
			str += '"';
		} else {
			str += value.ToString(sig, false);
		}
		str += " => ";
		str += iter->second.ToString(sig);
	}
	str += "}";
	return str;
}

const Value *Object_Dict::Find(Signal sig, const Value &valueIdx) const
{
	if (!ValueDict::IsValidKey(valueIdx)) {
		sig.SetError(ERR_KeyError, "invalid value type for key");
		return NULL;
	}
	ValueDict::const_iterator iter = _valDict.find(valueIdx);
	return (iter == _valDict.end())? NULL : &iter->second;
}

void Object_Dict::SetError_KeyNotFound(Signal sig, const Value &valueIdx)
{
	sig.SetError(ERR_KeyError, "dictionary doesn't have a key '%s'",
										valueIdx.ToString(sig).c_str());
}

//-----------------------------------------------------------------------------
// Object_Dict::IteratorKeys
//-----------------------------------------------------------------------------
Object_Dict::IteratorKeys::IteratorKeys(Object_Dict *pObj) :
				Iterator(false), _pObj(pObj), _pCur(pObj->GetDict().begin())
{
}

Object_Dict::IteratorKeys::~IteratorKeys()
{
	Object::Delete(_pObj);
}

bool Object_Dict::IteratorKeys::Next(Signal sig, Value &value)
{
	if (_pCur == _pObj->GetDict().end()) return false;
	value = _pCur->first;
	_pCur++;
	return true;
}

//-----------------------------------------------------------------------------
// Object_Dict::IteratorValues
//-----------------------------------------------------------------------------
Object_Dict::IteratorValues::IteratorValues(Object_Dict *pObj) :
				Iterator(false), _pObj(pObj), _pCur(pObj->GetDict().begin())
{
}

Object_Dict::IteratorValues::~IteratorValues()
{
	Object::Delete(_pObj);
}

bool Object_Dict::IteratorValues::Next(Signal sig, Value &value)
{
	if (_pCur == _pObj->GetDict().end()) return false;
	value = _pCur->second;
	_pCur++;
	return true;
}

//-----------------------------------------------------------------------------
// Object_Dict::IteratorPairs
//-----------------------------------------------------------------------------
Object_Dict::IteratorPairs::IteratorPairs(Object_Dict *pObj) :
				Iterator(false), _pObj(pObj), _pCur(pObj->GetDict().begin())
{
}

Object_Dict::IteratorPairs::~IteratorPairs()
{
	Object::Delete(_pObj);
}

bool Object_Dict::IteratorPairs::Next(Signal sig, Value &value)
{
	if (_pCur == _pObj->GetDict().end()) return false;
	ValueList &valList = value.InitAsList(*_pObj);
	valList.push_back(_pCur->first);
	valList.push_back(_pCur->second);
	_pCur++;
	return true;
}

//-----------------------------------------------------------------------------
// AScript Interface
//-----------------------------------------------------------------------------
// Constructor: Dict(elem[]?) {block?}
Object_Dict::Constructor::Constructor(Environment &env, const char *name) :
							ConstructorBase(name, env.GetClass_Dict())
{
	DeclareArg(env, "elem", VTYPE_AnyType, true, OCCUR_ZeroOrOnce);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

Value Object_Dict::Constructor::DoEval(Environment &env, Signal sig, Context &context) const
{
	Object_Dict *pObj = new Object_Dict(env.GetClass_Dict());
	Value result(pObj, VTYPE_Dict);
	ValueDict &valDict = pObj->GetDict();
	if (context.GetValue(0).IsList()) {
		if (!valDict.Store(sig, context.GetList(0))) return Value::Null;
	}
	if (context.IsBlockSpecified()) {
		Environment envLister(&env, ENVTYPE_Lister);
		Value valueList = context.GetBlock()->GetExprList().ExecForList(envLister, sig, false);
		if (sig.IsSignalled() || !valueList.IsList()) return Value::Null;
		if (!valDict.Store(sig, valueList.GetList())) return Value::Null;
	}
	return result;
}

// Dict#len()
AScript_DeclareMethod(Dict, len)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementMethod(Dict, len)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	return Value(static_cast<Number>(pSelf->GetDict().size()));
}

// Dict#get(key, default?):map:[raise]
AScript_DeclareMethod(Dict, get)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "key", VTYPE_AnyType);
	DeclareArg(env, "default", VTYPE_AnyType, false, OCCUR_ZeroOrOnce);
	DeclareAttr(AScript_Symbol(raise));
}

AScript_ImplementMethod(Dict, get)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	const Value &valueIdx = context.GetValue(0);
	const Value *pValue = pSelf->Find(sig, valueIdx);
	if (pValue != NULL) {
		return *pValue;
	} else if (context.IsSet(AScript_Symbol(raise))) {
		Object_Dict::SetError_KeyNotFound(sig, valueIdx);
		return Value::Null;
	} else {
		return context.GetValue(1);
	}
}

// Dict#has_key(key):map
AScript_DeclareMethod(Dict, has_key)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "key", VTYPE_AnyType);
}

AScript_ImplementMethod(Dict, has_key)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	const Value &valueIdx = context.GetValue(0);
	const Value *pValue = pSelf->Find(sig, valueIdx);
	return Value(pValue != NULL);
}

// Dict#keys() {block?}
AScript_DeclareMethod(Dict, keys)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(Dict, keys)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	Object_Dict *pObj = dynamic_cast<Object_Dict *>(pSelf->IncRef());
	return ReturnIterator(env, sig, context,
							new Object_Dict::IteratorKeys(pObj));
}

// Dict#values() {block?}
AScript_DeclareMethod(Dict, values)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(Dict, values)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	Object_Dict *pObj = dynamic_cast<Object_Dict *>(pSelf->IncRef());
	return ReturnIterator(env, sig, context,
							new Object_Dict::IteratorValues(pObj));
}

// Dict#pairs() {block?}
AScript_DeclareMethod(Dict, pairs)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareBlock(OCCUR_ZeroOrOnce);
}

AScript_ImplementMethod(Dict, pairs)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	Object_Dict *pObj = dynamic_cast<Object_Dict *>(pSelf->IncRef());
	return ReturnIterator(env, sig, context,
							new Object_Dict::IteratorPairs(pObj));
}

// Dict#clear!()
AScript_DeclareMethod(Dict, clear_X)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
}

AScript_ImplementMethod(Dict, clear_X)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	pSelf->GetDict().clear();
	return Value::Null;
}

// Dict#add!(elem[])
AScript_DeclareMethod(Dict, add_X)
{
	SetMode(RSLTMODE_Normal, MAP_Off, FLAT_Off);
	DeclareArg(env, "elem", VTYPE_AnyType, true);
}

AScript_ImplementMethod(Dict, add_X)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	if (!pSelf->GetDict().Store(sig, context.GetList(0))) return Value::Null;
	return Value::Null;
}

// Dict#erase!(key)
AScript_DeclareMethod(Dict, erase_X)
{
	SetMode(RSLTMODE_Normal, MAP_On, FLAT_Off);
	DeclareArg(env, "key", VTYPE_AnyType);
}

AScript_ImplementMethod(Dict, erase_X)
{
	Object_Dict *pSelf = Object_Dict::GetSelfObj(context);
	pSelf->GetDict().erase(context.GetValue(0));
	return Value::Null;
}

// Assignment
Class_Dict::Class_Dict(Class *pClassSuper, const Symbol *pSymbol) :
												Class(pClassSuper, pSymbol)
{
	AScript_AssignMethod(Dict, len);
	AScript_AssignMethod(Dict, get);
	AScript_AssignMethod(Dict, has_key);
	AScript_AssignMethod(Dict, keys);
	AScript_AssignMethod(Dict, values);
	AScript_AssignMethod(Dict, pairs);
	AScript_AssignMethodEx(Dict, clear_X, "clear!");
	AScript_AssignMethodEx(Dict, add_X, "add!");
	AScript_AssignMethodEx(Dict, erase_X, "erase!");
}

Object *Class_Dict::CreateDescendant(Environment &env, Signal sig, Class *pClass)
{
	return new Object_Dict((pClass == NULL)? this : pClass);
}

}
